/**
 * Copyright (c) 2008 Aria Stewart
 * Based on code Copyright (c) 2005 Zed A. Shaw
 * You can redistribute it and/or modify it under the same terms as Ruby.
 */
#import "HttpParser.h"
#import "HttpRequest.h"
#import "Foundation/Foundation.h"
#include <stdio.h>
#include <assert.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>

@implementation HttpParser
  int cs;
  size_t body_start;
  int content_len;
  size_t nread;
  size_t mark;
  size_t field_start;
  size_t field_len;
  size_t query_start;
	id handler_map;
	HttpRequest *request;

#define LEN(AT, FPC) (FPC - buffer - AT)
#define MARK(M,FPC) (M = (FPC) - buffer)
#define PTR_TO(F) (buffer + F)

#define TRIE_INCREASE 30

/** Machine **/

%%{
  
  machine http_parser;

  action mark {MARK(mark, fpc); }

  action start_field { MARK(field_start, fpc); }
  action write_field { 
    field_len = LEN(field_start, fpc);
  }

  action start_value { MARK(mark, fpc); }
  action write_value { 
		[self setHttpField: [[NSString alloc] initWithCString: PTR_TO(field_start) 
				length: field_len] 
			value: [[NSString alloc] initWithCString: PTR_TO(mark) 
				length: LEN(mark, fpc)]];
  }
  action request_method { 
    [self setRequestMethod: [[NSString alloc] initWithCString:PTR_TO(mark) length:LEN(mark, fpc)]];
  }
  action request_uri { 
		[self setRequestUri: [[NSString alloc] initWithCString:PTR_TO(mark) length:LEN(mark, fpc)]];
  }
  action fragment { 
    [self setFragment: [[NSString alloc] initWithCString:PTR_TO(mark) length:LEN(mark, fpc)]];
  }

  action start_query {MARK(query_start, fpc); }
  action query_string { 
		[self setQueryString: [[NSString alloc] initWithCString: PTR_TO(query_start) length: LEN(query_start, fpc)]];
  }

  action http_version {	
		[self setHttpVersion: [[NSString alloc] initWithCString: PTR_TO(mark) length: LEN(mark, fpc)]];
  }

  action request_path {
		[self setRequestPath: [[NSString alloc] initWithCString: PTR_TO(mark) length: LEN(mark,fpc)]];
  }

  action done { 
    body_start = fpc - buffer + 1; 
    [self headerDone: [[NSString alloc] initWithCString: fpc + 1 length: pe - fpc - 1]];
    fbreak;
  }

  include http_parser_common "http11_parser_common.rl";

}%%

/** Data **/
%% write data;

/** exec **/



/** Defines common length and error messages for input length validation. */
#define DEF_MAX_LENGTH(N,length) const size_t MAX_##N##_LENGTH = length; NSString *MAX_##N##_LENGTH_ERR = @"HTTP element " # N  " is longer than the " # length " allowed length."

/** Validates the max length of given input and throws an HttpParserError exception if over. */
#define VALIDATE_MAX_LENGTH(len, N) if(len > MAX_##N##_LENGTH) { [NSException raise: @"HttpParserError" format: MAX_##N##_LENGTH_ERR] ; }

/* Defines the maximum allowed lengths for various input elements.*/
DEF_MAX_LENGTH(FIELD_NAME, 256);
DEF_MAX_LENGTH(FIELD_VALUE, 80 * 1024);
DEF_MAX_LENGTH(REQUEST_URI, 1024 * 12);
DEF_MAX_LENGTH(FRAGMENT, 1024); /* Don't know if this length is specified somewhere or not */
DEF_MAX_LENGTH(REQUEST_PATH, 1024);
DEF_MAX_LENGTH(QUERY_STRING, (1024 * 10));
DEF_MAX_LENGTH(HEADER, (1024 * (80 + 32)));

-(void) setHttpField: (NSString *)field value:(NSString *)value 
{
  VALIDATE_MAX_LENGTH([field length], FIELD_NAME);
  VALIDATE_MAX_LENGTH([value length], FIELD_VALUE);

	[request setValue: value forKey: [NSString stringWithFormat: @"HTTP_%@",
		[field uppercaseString]]];
}

-(void) setRequestMethod: (NSString *)val
{
	[request setValue: val forKey: @"REQUEST_METHOD"];
}

-(void) setRequestUri: (NSString *)val
{
	VALIDATE_MAX_LENGTH([val length], REQUEST_URI);
	[request setValue: val forKey: @"REQUEST_URI"];
}

-(void) setFragment: (NSString *)val
{
	VALIDATE_MAX_LENGTH([val length], FRAGMENT);
	[request setValue: val forKey: @"FRAGMENT"];
}

-(void) setRequestPath: (NSString *)val
{
	VALIDATE_MAX_LENGTH([val length], REQUEST_PATH);
	[request setValue: val forKey: @"REQUEST_PATH"];
}

-(void) setQueryString: (NSString *)val
{
	VALIDATE_MAX_LENGTH([val length], QUERY_STRING);
	[request setValue: val forKey: @"QUERY_STRING"];
}

-(void) setHttpVersion: (NSString *)val
{
	[request setValue: val forKey: @"HTTP_VERSION"];
}

/** Finalizes the request header to have a bunch of stuff that's
	needed. */

-(void) headerDone: (NSString *) initial_http_body
{
	id temp = nil;
	id ctype = nil;
	id clen = nil;
	NSRange colon;

	clen = [request valueForKey: @"HTTP_CONTENT_LENGTH"];
	if(clen != nil) {
		[request setValue: clen forKey: @"CONTENT_LENGTH"];
	}

	ctype = [request valueForKey: @"HTTP_CONTENT_TYPE"];
	if(ctype != nil) {
		[request setValue: ctype forKey: @"CONTENT_TYPE"];
	}

	[request setValue: @"CGI/1.2" forKey: @"GATEWAY_INTERFACE"];
	if((temp = [request valueForKey: @"HTTP_HOST"]) != nil) {
		colon = [temp rangeOfString: @":"];
		if(colon.location != NSNotFound) {
			[request setValue: [temp substringToIndex: colon.location] forKey: @"SERVER_NAME"];
			[request setValue: [temp substringFromIndex: colon.location + 1] forKey: @"SERVER_PORT"]; 
		} else {
			[request setValue: temp forKey: @"SERVER_NAME"];
			[request setValue: @"80" forKey: @"SERVER_PORT"];
		}
	}

	/* grab the initial body and stuff it into an ivar */
	[request setInitialHttpBody: initial_http_body];
  [request setValue: @"HTTP/1.1" forKey: @"SERVER_PROTOCOL"];
  [request setValue: @"Mongrel 1.1.3" forKey: @"SERVER_SOFTWARE"];
}


/**
 *    [parser reset];
 *
 * Resets the parser to it's initial state so that you can reuse it
 * rather than making new ones.
 */
-(void) reset
{
  [self init];
}


/**
 *   [parser finish] -> BOOL
 *
 * Finishes a parser early which could put in a "good" or bad state.
 * You should call reset after finish or bad things will happen.
 */
-(BOOL) finish
{
  %%write eof;

  if ([self has_error]) {
    return NO;
  } else if ([self is_finished]) {
    return YES;
  } else {
    return NO;
  }
}

/**
 * call-seq:
 *    [parser execute: data] -> Integer
 *
 * calls [parser execute: data start: 0]
 */
-(size_t) execute: (NSData *) data
{
	return [self execute: data start: 0];
}

/**
 * call-seq:
 *    [parser execute: data start:start] -> Integer
 *
 * Takes a String of data, parses the String of data filling in
 * self, returning an Integer to indicate how much of the data has been
 * read.  No matter what the return value, you should call
 * HttpParser#is_finished and HttpParser#has_error to figure out if it's done
 * parsing or there was an error.
 * 
 * This function throws an exception when there is a parsing error.
 * This makes the logic for working with the parser much easier.  You can
 * still test for an error, but you need to wrap the parser with an
 * exception handling block.
 *
 * The second argument allows for parsing a partial request and then
 * continuing the parsing from that position.  It needs all of the original
 * data as well so you have to append to the data buffer as you read.
 */
-(size_t)execute: (NSData *)sdata start:(int)start {

  if(start >= [sdata length]) {
		[NSException raise: @"HttpParserError" format: @"Requested start is after data buffer end"];
  } else {
		{
			const char *buffer = [sdata bytes];
			size_t len = [sdata length];
			size_t off = start;
			const char *p, *pe;

			assert(off <= len && "offset past end of buffer");

			p = buffer+off;
			pe = buffer+len;

			assert(pe - p == len - off && "pointers aren't same distance");

			%% write exec;

			nread += p - (buffer + off);

			assert(p <= pe && "buffer overflow after parsing execute");
			assert(nread <= len && "nread longer than length");
			assert(body_start <= len && "body starts after buffer end");
			assert(mark < len && "mark is after buffer end");
			assert(field_len <= len && "field has length longer than whole buffer");
			assert(field_start < len && "field starts after buffer end");

			if(body_start) {
				/* final \r\n combo encountered so stop right here */
				%%write eof;
				nread++;
			}
		}

    VALIDATE_MAX_LENGTH([self nread], HEADER);

    if([self has_error]) {
			[NSException raise: @"HttpParserError" format: @"Invalid HTTP format, parsing fails."];
    } else {
      return [self nread];
    }
  }
}

/**
 *    [parser has_error] -> BOOL
 *
 * Tells you whether the parser is in an error state.
 */
-(BOOL) has_error
{
  return cs == http_parser_error ? YES : NO;
}

/**
 *    [parser nread] -> int
 *
 * Returns the amount of data processed so far during this processing cycle.
 * It is set to 0 on initialize or reset calls and is incremented each time
 * execute is called.
 */
-(int) nread
{
	return nread;
}

/**
 *   [parser is_finished] -> BOOL
 *
 * Tells you whether the parser is finished or not and in a good state.
 */
-(BOOL) is_finished
{
  return cs == http_parser_first_final;
}

-(id) init
{
	request = [[HttpRequest alloc] init];
	body_start = 0;
	cs = 0;
	nread = 0;
	content_len = 0;
	mark = 0;
	field_start = 0;
	field_len = 0;
	query_start = 0;
	handler_map = nil;

  %% write init;
  return self;
}

-(id) requestObject
{
	return request;
}

@end
